/*
 * Copyright (C) 2006  Emmanuel VARAGNAT <>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <getopt.h>

#include "common.h"
#include "scan.h"
#include "array.h"

#define DEBUG_MODE

/*
 *  Yes it's a GCC extension, but who cares ?
 */
#ifdef DEBUG_MODE
# define DEBUG(s, ...) printf(s, __VA_ARGS__)
#else
# define DEBUG(s, ...)
#endif

#define RET             "\n"

struct common_boot_sector {
  uint8_t   jmp_instr[3];
  char      oem_name[8];
  uint16_t  byte_per_sector;
  uint8_t   sector_per_cluster;
  uint16_t  reserved_area_size;  /* in number of sector */
  uint8_t   num_fat;
  uint16_t  root_dir_max_files;  /* 0 for FAT32 */
  uint16_t  num_sector_pt1;
  uint8_t   media_type;
  uint16_t  fat_size;            /* in sector; 0 for FAT32 */
  uint16_t  sector_per_track;
  uint16_t  number_of_head;
  uint32_t  start_of_partition;
  uint32_t  num_sector_pt2;
} __attribute__((packed));


struct fat12_boot_sector {
  uint8_t  bios_drive_num;
  uint8_t  unused1;
  uint8_t  boot_signature_ext; /* equal to 0x29 */
  uint32_t serial_number;
  char     volume_label[11];
  char     type_label[8];
  uint8_t  unused2[448];
  uint16_t magic_number;  /* 0xAA55 */
} __attribute__((packed));


struct fat32_boot_sector {
  uint32_t fat_size;      /* in sector */
  uint16_t multi_fat_desc;
  uint16_t major_minor_version;
  uint32_t root_dir_cluster;
  uint16_t fsinfo_sector;
  uint16_t bootsect_backup_sector;
  uint8_t  reserved[12];

  /* FAT12/FAT16 like */
  uint8_t  bios_drive_num;
  uint8_t  unused1;
  uint8_t  boot_signature_ext; /* equal to 0x29 */
  uint32_t serial_number;
  char     volume_label[11];
  char     type_label[8];

  uint8_t  unused2[420];
  uint16_t magic_number;  /* 0xAA55 */  
} __attribute__((packed));


struct boot_sector {
  struct common_boot_sector common;
  union {
    struct fat12_boot_sector fat12;
    struct fat12_boot_sector fat32;    
  };
} __attribute__((packed));

struct dir_entry {
  char padding[32];
} __attribute__((packed));

static struct array EXIF_hdr_offset;
static struct array JFIF_hdr_offset;

static unsigned char *fat_bitmap;
static uint64_t num_data_cluster;

static uint64_t start_sector;
uint32_t (*next_cluster)(int fd,
			 uint32_t cluster,
			 const struct boot_sector *bsect);
static int flags;

void EXIF_found(off_t offset, void* extra);
void JFIF_found(off_t offset, void* extra);

struct motif_infos jpeg_hdr_motifs[] = {
  { 4, { 0xff, 0xd8, 0xff, 0xe1 }, EXIF_found, &EXIF_hdr_offset }, /* SOI followed by EXIF */ 
  { 4, { 0xff, 0xd8, 0xff, 0xe0 }, JFIF_found, &JFIF_hdr_offset }, /* SOI followed by JFIF */ 
  { 0, {}, NULL, NULL }
};

void boot_sector_display_infos(const struct boot_sector *bsect) {
  printf("IMPORTANT INFORMATIONS"                  RET
	 "Byte per sector:           %u"           RET
	 "Sector per cluster:        %u"           RET
	 "Reserved area size:        %u cluster(s)"  RET
	 "Number of FAT:             %u"           RET
	 "Max number of files"                     RET
	 "              in root dir: %u"           RET
	 "FAT size:                  %u sector(s)" RET
	 "Number of sectors:         %u"           RET,
	 bsect->common.byte_per_sector,
	 bsect->common.sector_per_cluster,
	 bsect->common.reserved_area_size,
	 bsect->common.num_fat	, 
	 bsect->common.root_dir_max_files,
	 bsect->common.fat_size,	 
	 ((bsect->common.num_sector_pt1 == 0) ?
	  bsect->common.num_sector_pt2 : (uint32_t)bsect->common.num_sector_pt1)
	 );

  printf(RET);

  printf("OEM name:           %.8s\n",
	 bsect->common.oem_name);
  
}

void EXIF_found(off_t offset, void* extra) {
  DEBUG("EXIF signature found at %lu [%u] [230]=%lu\n", offset, ((struct array*)extra)->last, ((off_t*)((struct array*)extra)->data)[230]);

  if( offset % SECTOR_SIZE )
    fprintf(stderr, "Non aligned EXIF JPEG header found\n");
  else
    array_append((struct array*)extra, &offset);
}

void JFIF_found(off_t offset, void* extra) {
  DEBUG("JFIF signature found at %lu\n", offset);

  if( offset % SECTOR_SIZE )
    fprintf(stderr, "Non aligned JFIF JPEG header found\n");
  else
    array_append((struct array*)extra, &offset);
}

void dump_cluster(int jfd,
		  int bfd,
		  uint32_t cluster,
		  const struct boot_sector *bsect)
{
  off_t cluster2, read_offset;
  size_t cluster_size;
  char *buf;

  DEBUG("Dump cluster : %u\n", cluster);

  /*
   * FIXME: computed each time
   */
  cluster_size = bsect->common.sector_per_cluster * bsect->common.byte_per_sector;

  DEBUG(" cluster size : %u\n", cluster_size);

  cluster2 =
    ( bsect->common.reserved_area_size
      + bsect->common.fat_size * bsect->common.num_fat
      + (bsect->common.root_dir_max_files * sizeof(struct dir_entry)) / bsect->common.byte_per_sector
     )
    * bsect->common.byte_per_sector;

  DEBUG(" cluster2 offset : %lu\n", cluster2);

  read_offset =
    start_sector * SECTOR_SIZE
    + cluster2
    + (cluster - 2) * cluster_size;

  DEBUG(" read_offset : %lu\n", read_offset);
  DEBUG(" jfd=%d  bfd=%d\n", jfd, bfd);

  buf = alloca(cluster_size);

  assert( fat_bitmap[cluster] < 255 );
  if(cluster < num_data_cluster)
    fat_bitmap[cluster]++;
  else
    printf("Dbordement: %u >= %u\n", cluster, num_data_cluster);
  
  if( lseek(bfd, read_offset, SEEK_SET) == -1 )
    INTERNAL_ERROR_EXIT("lseek: ", strerror(errno));

  if( read(bfd, buf, cluster_size) == -1 )
    INTERNAL_ERROR_EXIT("read: ", strerror(errno));

  if(! (flags & FLAGS_DRY_RUN))
    if( write(jfd, buf, cluster_size) == -1 )
      INTERNAL_ERROR_EXIT("write: ", strerror(errno));
}

uint32_t next_cluster_12(int fd,
			 uint32_t cluster,
			 const struct boot_sector *bsect)
{
  unsigned char buf[2];
  uint16_t next;
  off_t offset;  

  offset =
    (start_sector + bsect->common.reserved_area_size) * SECTOR_SIZE
    + (3 * cluster) / 2;

  if( lseek(fd, offset, SEEK_SET) == (off_t)-1 )
    INTERNAL_ERROR_EXIT("lseek: ", strerror(errno));

  if( read(fd, buf, 2) == -1 )
    INTERNAL_ERROR_EXIT("read: ", strerror(errno));

  printf("offset: %ld buf[0]=0x%x buf[1]=0x%x\n", offset, buf[0], buf[1]);
  if (cluster % 2) {
    next = (int)buf[1];
    printf("0: 0x%x\n", next);
    next = next << 8;
    printf("1: 0x%x\n", next);
    next += buf[0];
    printf("2: 0x%x\n", next);
    next = next >> 4;
    printf("3: 0x%x\n", next);
  }
  else {
    next = (int)buf[1];
    printf("0: 0x%x\n", next);
    next = next & 0x0f;
    printf("1: 0x%x\n", next);
    next = next << 8;
    printf("2: 0x%x\n", next);
    next += buf[0];
    printf("3: 0x%x\n", next);
    //    next = (int)(buf[1] & 0x0f) << 8 + (int)buf[0];
  }

  return ((next == 0x0fff) ? (uint32_t)-1 : (uint32_t)next);
}

uint32_t next_cluster_16(int fd,
			 uint32_t cluster,
			 const struct boot_sector *bsect)
{
  char buf[2];
  uint16_t next;
  off_t offset;

  offset =
    (start_sector + bsect->common.reserved_area_size) * SECTOR_SIZE
    + 2 * cluster;

  if( lseek(fd, offset, SEEK_SET) == (off_t)-1 )
    INTERNAL_ERROR_EXIT("lseek: ", strerror(errno));

  if( read(fd, buf, 2) == -1 )
    INTERNAL_ERROR_EXIT("read: ", strerror(errno));

  memcpy(&next, buf, 2);

  return ((next == 0xffff) ? (uint32_t)-1 : (uint32_t)next);
}


void dump_file(int fd,
	       uint32_t first_cluster,
	       const struct boot_sector *bsect,
	       const char *destdir)
{
  uint32_t cluster;
  int jfd = -1;
  int len;
  char *basename = "jpeg.recover.XXXXXX";
  static char path1[MAXPATHLEN];
  static char path2[MAXPATHLEN];

  if(! (flags & FLAGS_DRY_RUN)) {
    len = strlen(basename)
      + strlen(destdir) /* it's lenght should have been tested before */
      + 6;              /* '.jpg' extension + '\0' and '/' */
    
    if(len > MAXPATHLEN) {
      fprintf(stderr, "pathname too long\n");
      return;
    }
    
    strcpy(path1, destdir);
    strcat(path1, "/");
    strcat(path1, basename);
    
    jfd = mkstemp(path1);
    if(jfd == -1)
      INTERNAL_ERROR_EXIT("mkstemp: ", strerror(errno));
  
    strcpy(path2, path1);
    strcat(path2, ".jpg");
    
    DEBUG("  -> %s\n", path2);
    
    if( rename(path1, path2) == -1 )
      INTERNAL_ERROR_EXIT("rename: ", strerror(errno));
  }

  cluster = first_cluster;
  do {
    dump_cluster(jfd, fd, cluster, bsect);

    cluster = next_cluster(fd, cluster, bsect);
  } while(cluster != (uint32_t)-1 );

  if(! (flags & FLAGS_DRY_RUN))
    close(jfd);
}

int recover_photos(const char *filename, const char* destdir, int _flags) {
  struct boot_sector bsect;
  int fd;
  unsigned int i;

  flags = _flags;

  fd = open(filename, O_RDONLY);
  if(fd == -1) {
    perror("open");
    exit(1);
  }

  /* Search for boot sector */
  {
    int found, n, end;

    for (n = 0, found = 0, end = 0; ! found && ! end; n++) {

      switch( read(fd, &bsect, sizeof(struct boot_sector))) {
      case -1:
	INTERNAL_ERROR_EXIT("read: ", strerror(errno));
      case 0:
	end = 1;
	break;
      default:
	found = ( (bsect.fat32.magic_number == 0xAA55) &&
		  (bsect.fat32.boot_signature_ext == 0x29 || bsect.fat12.boot_signature_ext == 0x29) );
      }
    }

    if(found) {
      start_sector = --n;    /* We looped one time more */
      printf("Start sector: %d\n", start_sector);
    }
    else {
      printf("WARNING: no FAT boot sector found\n");
      exit(0);
    }
  }

  if( lseek(fd, start_sector * SECTOR_SIZE, SEEK_SET) == (off_t)-1 ) {
    perror("lseek");
    exit(1);
  }

  if( read(fd, &bsect, sizeof(struct boot_sector)) == -1 ) {
    perror("read");
    exit(1);
  }

  boot_sector_display_infos(&bsect);

  if( bsect.common.root_dir_max_files == 0 &&
      bsect.common.num_sector_pt1 == 0 &&
      bsect.fat32.boot_signature_ext == 0x29)
    {
      printf("Don't handle FAT32\n");
      exit(2);
    }

  /*
   *  Find first sector of data (cluster 2)
   */
  printf("Root directory offset: %u bytes\n",
	 bsect.common.byte_per_sector * 
	 ( bsect.common.sector_per_cluster * bsect.common.reserved_area_size
	   + bsect.common.fat_size * bsect.common.num_fat )
	 );

  printf("Root directory offset: sector %u\n",
	 bsect.common.reserved_area_size
	   + bsect.common.fat_size * bsect.common.num_fat);

  printf("Cluster 2 offset:      sector %u\n",
	 bsect.common.reserved_area_size
	 + bsect.common.fat_size * bsect.common.num_fat
	 + (bsect.common.root_dir_max_files * sizeof(struct dir_entry)) /  bsect.common.byte_per_sector
	 );
  printf("FAT start offset:      %d\n", (start_sector + bsect.common.reserved_area_size) * SECTOR_SIZE);

  if( bsect.common.num_fat == 2 ) {

    for(i = 0; i < bsect.common.fat_size; i++) {
      char sector1[SECTOR_SIZE];
      char sector2[SECTOR_SIZE];
      
      if( lseek(fd, (start_sector + bsect.common.reserved_area_size + i) * SECTOR_SIZE, SEEK_SET) == (off_t)-1 ) {
	perror("lseek");
	exit(1);
      }
      if( read(fd, sector1, SECTOR_SIZE) == -1 ) {
	perror("read");
	exit(1);
      }

      if( lseek(fd, (start_sector + bsect.common.reserved_area_size + i + bsect.common.fat_size) * SECTOR_SIZE, SEEK_SET) == (off_t)-1 ) {
	perror("lseek");
	exit(1);
      }
      if( read(fd, sector2, SECTOR_SIZE) == -1 ) {
	perror("read");
	exit(1);
      }

      if( bcmp(sector1, sector2, SECTOR_SIZE) != 0 )
	break;
    }

    printf("\n");
    if(i < bsect.common.fat_size) {
      printf("FATs are differents !!!\n");
      exit(1);
    }
    else
      printf("FATs are equals !!!\n");
  }

  /*
   * Determine number of entry in FAT and size of each entry
   *
   * FIXME: computed values are only good for FAT12/FAT16
   */
  {
    uint64_t num_sectors; /* total number of sectors :
			     reserved area + fat area + data area */
    short fat_entry_size;
    
    num_sectors = ((bsect.common.num_sector_pt1 == 0) ?
		   bsect.common.num_sector_pt2 : (uint32_t)bsect.common.num_sector_pt1);

    num_data_cluster =
      ( num_sectors
	- (bsect.common.reserved_area_size
	   + bsect.common.fat_size * bsect.common.num_fat) * bsect.common.byte_per_sector
	- 1
	) / bsect.common.sector_per_cluster;

    if( num_data_cluster < 4085 ) {
      printf("FAT type:\tFAT12\n");
      next_cluster = next_cluster_12;
      fat_entry_size = 12;
    }
    else if( num_data_cluster < 65525 ) {
      printf("FAT type:\tFAT16\n");
      next_cluster = next_cluster_16;
      fat_entry_size = 16;
    }
    else { /* as FAT32 is not supported... */
      fprintf(stderr, "Can't determine FAT type\n");
      exit(2);
    }
    
    fat_bitmap = calloc(num_data_cluster, 1);
  }
  
  {
    size_t ii;
    size_t cluster_size = bsect.common.byte_per_sector * bsect.common.sector_per_cluster;
    off_t cluster2;

    array_init(&EXIF_hdr_offset, sizeof(off_t), 300);
    array_init(&JFIF_hdr_offset, sizeof(off_t), 300);
    
    scan_for_motifs(fd, jpeg_hdr_motifs);

    cluster2 =
      start_sector
      + bsect.common.reserved_area_size
      + bsect.common.fat_size * bsect.common.num_fat
      + (bsect.common.root_dir_max_files * sizeof(struct dir_entry)) /  bsect.common.byte_per_sector;

    DEBUG("Dumping %u JPEG EXIF files found\n", EXIF_hdr_offset.last);
    
    for(ii = 0; ii < EXIF_hdr_offset.last; ii++) {
      off_t off = ((off_t*)EXIF_hdr_offset.data)[ii];
      off_t hdr_off = off - (cluster2 * SECTOR_SIZE);  /* FIXME: only good for FAT12/16 */

      DEBUG(" dumping EXIF file [%u] starting at %lu (cluster:%lu [modulo=%lu])\n",
	    ii,
	    off,
	    hdr_off / cluster_size,
	    hdr_off % cluster_size);

      dump_file(fd, hdr_off / cluster_size + 2, &bsect, destdir);
    }

    DEBUG("Dumping %u JPEG JFIF files found\n", EXIF_hdr_offset.last);
    for(ii = 0; ii < JFIF_hdr_offset.last; ii++) {
      off_t off = ((off_t*)JFIF_hdr_offset.data)[ii];
      off_t hdr_off = off - (cluster2 * SECTOR_SIZE);  /* FIXME: only good for FAT12/16 */

      DEBUG(" dumping JFIF file starting at %lu (cluster:%lu [modulo=%lu])\n",
	    off,
	    hdr_off / cluster_size,
	    hdr_off % cluster_size);

      dump_file(fd, hdr_off / cluster_size + 2, &bsect, destdir);
    }
  }

  {
    uint32_t dumped = 0;

    for(i = 0; i < num_data_cluster; i++) {
      if(fat_bitmap[i] >= 1) {
	dumped++;
	if(fat_bitmap[i] > 1)
	  printf("INFO: cluster %u dumped %u times\n", i, fat_bitmap[i]);
      }
    }

    printf("%u over %llu cluster dumped\n", dumped, num_data_cluster);
    printf("Clusters assigned to directories and free clusters are not\n"
	   "taken into account\n");
  }

  return 0;
}
